from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Literal, TypeAlias, TypedDict

from ..sparql import Binding, as_long, as_uri, as_opt_uri, as_string, as_datetime, as_opt_float
from ..metacore import UriResource
from ..geofeaturemeta import Point


@dataclass(frozen=True)
class DataObjectLite:
	"""
	Dataclass for basic metadata of a data object.

	Attributes:
		`uri` (string): landing page URI
		`filename` (string): file name
		`size_bytes` (int): size of the data object in bytes
		`datatype_uri` (string): URI of the data type of this data
			object, internally referred to as data object spec
		`station_uri` (string | None): URI of the station that acquired
			the data (if applicable)
		`sampling_height` (float | None): sampling height in meters, if
			applicable
		`submission_time` (datetime): time stamp of the end of data
			submission
		`time_start` (datetime): time stamp of the beginning of the
			object's temporal coverage
		`time_end` (datetime): time stamp of the end of the object's
			temporal coverage
	"""
	uri: str
	filename: str
	size_bytes: int
	datatype_uri: str
	station_uri: str | None
	sampling_height: float | None
	submission_time: datetime
	time_start: datetime
	time_end: datetime

CategorySelector: TypeAlias = str | UriResource | list[str] | list[UriResource] | None
TemporalProp = Literal["submTime", "timeStart", "timeEnd"]
IntegerProp = Literal["size"]
StringProp = Literal["fileName"]
OrderByProp: TypeAlias = TemporalProp | IntegerProp | StringProp
ExclusiveComparator: TypeAlias = Literal["<", ">"]
Comparator: TypeAlias = ExclusiveComparator | Literal["<=", ">=", "="]

class Filter(ABC):
	@abstractmethod
	def render(self) -> str: pass

@dataclass(frozen=True)
class TimeFilter(Filter):
	prop: TemporalProp
	op: ExclusiveComparator
	value: datetime | str

	def render(self) -> str:
		v = self.value
		dtStr = v.isoformat() if type(v) is datetime else str(v)
		return f'?{self.prop} {self.op} "{dtStr}"^^xsd:dateTime'

@dataclass(frozen=True)
class SizeFilter(Filter):
	op: Comparator
	value: int

	def render(self) -> str:
		return f'?size {self.op} "{self.value}"^^xsd:integer'

@dataclass(frozen=True)
class SamplingHeightFilter(Filter):
	op: Comparator
	value: float

	def render(self) -> str:
		return f'?samplingHeight {self.op} "{self.value}"^^xsd:float'

def polygon_to_wkt(poly: list[Point]) -> str:
	full_poly = poly
	if len(poly) >= 3 and poly[0] != poly[-1]:
		full_poly = list(poly)
		full_poly.append(poly[0])
	coord_list = ", ".join([f"{p.lon} {p.lat}" for p in full_poly])
	return f"POLYGON(({coord_list}))"

def box_intersect(min: Point, max: Point) -> 'GeoIntersectFilter':
	assert min.lat < max.lat and min.lon < max.lon, "Lat/lon region-of-interest box must have positive area"
	poly = [min, Point(max.lat, min.lon), max, Point(min.lat, max.lon), min]
	return GeoIntersectFilter(poly)

@dataclass(frozen=True)
class GeoIntersectFilter(Filter):
	polygon: list[Point]

	def render(self) -> str:
		assert len(self.polygon) >= 3, "polygon must have at least 3 points"
		return f'geo:sfIntersects/geo:asWKT "{polygon_to_wkt(self.polygon)}"^^geo:wktLiteral .'

class SortProp(TypedDict):
	prop: OrderByProp
class OrderBy(SortProp, total = False):
	descending: bool

def _order_clause(order: OrderBy | OrderByProp | None) -> str:
	if order is None:
		return ""
	if type(order) is str:
		return f"order by ?{order}\n"
	if type(order) is dict:
		prop = order['prop']
		return f"order by desc(?{prop})\n" if order.get("descending") else f"order by ?{prop}\n"
	else:
		return ""

def _selector_values_clause(varname: str, uris: list[str]) -> str:
	if len(uris) == 0:
		return ""
	else:
		return f"VALUES ?{varname} {{ {'<' + '> <'.join(uris) + '>'} }}"

def _get_uri_list(selector: CategorySelector) -> list[str]:
	if selector is None: return []
	def to_uri(uri_or_res: object) -> str:
		return getattr(uri_or_res, "uri", None) or str(uri_or_res)
	if type(selector) is list:
		return [to_uri(uri_or_res) for uri_or_res in selector]
	return [to_uri(selector)]

def dataobj_lite_list(
	datatype: CategorySelector,
	station: CategorySelector,
	filters: list[Filter],
	include_deprecated: bool,
	order_by: OrderBy | OrderByProp | None,
	limit: int,
	offset: int
) -> str:

	specUris =  _get_uri_list(datatype)
	stationUris = _get_uri_list(station)
	stationMandatory = "?dobj cpmeta:wasAcquiredBy/prov:wasAssociatedWith ?station ."
	stationLink = f"OPTIONAL{{ {stationMandatory} }}" if len(stationUris) == 0 else stationMandatory
	samplingHeightPresent = any(f is SamplingHeightFilter for f in filters)
	heightMandatory = "?dobj cpmeta:wasAcquiredBy/cpmeta:hasSamplingHeight ?samplingHeight ."
	heightLink = heightMandatory if samplingHeightPresent else  f"OPTIONAL{{ {heightMandatory} }}"

	non_geo_filters = [f for f in filters if not isinstance(f, GeoIntersectFilter)]
	geo_filters = [f for f in filters if isinstance(f, GeoIntersectFilter)]

	filter_clauses = "" if len(non_geo_filters) == 0 else "FILTER(" + " && ".join([
							f.render() for f in non_geo_filters
						]) + ")"
	geo_clause = "" if len(geo_filters) == 0 else "\n".join(["?dobj " + f.render() for f in geo_filters])

	geo_prefix = "prefix geo: <http://www.opengis.net/ont/geosparql#>" if len(geo_filters) > 0 else ""

	return f"""
prefix cpmeta: <http://meta.icos-cp.eu/ontologies/cpmeta/>
prefix prov: <http://www.w3.org/ns/prov#>
prefix xsd: <http://www.w3.org/2001/XMLSchema#>
{geo_prefix}

select ?dobj ?spec ?station ?samplingHeight ?fileName ?size ?submTime ?timeStart ?timeEnd
where {{
	{_selector_values_clause('spec', specUris)}
	?dobj cpmeta:hasObjectSpec ?spec .
	{_selector_values_clause('station', stationUris)}
	{stationLink}
	{heightLink}
	?dobj cpmeta:hasSizeInBytes ?size .
	?dobj cpmeta:hasName ?fileName .
	?dobj cpmeta:wasSubmittedBy/prov:endedAtTime ?submTime .
	?dobj cpmeta:hasStartTime | (cpmeta:wasAcquiredBy / prov:startedAtTime) ?timeStart .
	?dobj cpmeta:hasEndTime | (cpmeta:wasAcquiredBy / prov:endedAtTime) ?timeEnd .
	{"" if include_deprecated else "FILTER NOT EXISTS {[] cpmeta:isNextVersionOf ?dobj}"}
	{filter_clauses}
	{geo_clause}
}}
{_order_clause(order_by)}offset {offset} limit {min(limit, 10000)}"""

def parse_dobj_lite(row: Binding) -> DataObjectLite:
	return DataObjectLite(
		uri = as_uri("dobj", row),
		filename = as_string("fileName", row),
		size_bytes = as_long("size", row),
		datatype_uri = as_uri("spec", row),
		station_uri = as_opt_uri("station", row),
		sampling_height = as_opt_float("samplingHeight", row),
		submission_time = as_datetime("submTime", row),
		time_start = as_datetime("timeStart", row),
		time_end = as_datetime("timeEnd", row)
	)



import os
from .envri import EnvriConfig
from .auth import AuthTokenProvider, ConfigFileAuth, PasswordAuth, TokenAuth
from .metaclient import MetadataClient
from .dataclient import DataClient
from typing import Tuple

class Bootstrap():
	def __init__(self, conf: EnvriConfig) -> None:
		self._conf = conf
		data_path_var = "PORTAL_DATA_PATH_" + conf.envri.upper()
		data_path = os.environ.get(data_path_var)
		self._data_path = data_path if data_path is not None and len(data_path) > 0 else None


	def fromPasswordFile(
		self, conf_file_path: str | None = None
	) -> Tuple[ConfigFileAuth, MetadataClient, DataClient]:
		auth = ConfigFileAuth(self._conf, conf_file_path)
		data = DataClient(self._conf, auth, self._data_path)
		return auth, data.meta, data

	def fromCookieToken(self, token: str) -> Tuple[MetadataClient, DataClient]:
		auth = TokenAuth(token)
		data = DataClient(self._conf, auth, self._data_path)
		return data.meta, data

	def fromCredentials(self, user_id: str, password: str) -> Tuple[MetadataClient, DataClient]:
		auth = PasswordAuth(user_id, password, self._conf)
		data = DataClient(self._conf, auth, self._data_path)
		return data.meta, data

	def fromAuthProvider(self, auth: AuthTokenProvider) -> DataClient:
		return DataClient(self._conf, auth, self._data_path)



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Literal, Type, TypeAlias, TypeVar, Any
from dacite import Config, from_dict
import json

Orcid: TypeAlias = str
DoiMeta: TypeAlias = object
Instant: TypeAlias = str
URI: TypeAlias = str
CountryCode: TypeAlias = str
Sha256Sum: TypeAlias = str
LocalDate: TypeAlias = str
JsValue: TypeAlias = object

PinKind: TypeAlias = Literal["Sensor" , "Other"]

@dataclass(frozen=True)
class FeatureCollection:
	features: list[GeoFeature]
	label: Optional[str]
	uri: Optional[URI]


@dataclass(frozen=True)
class Position:
	lat: float
	lon: float
	alt: Optional[float]
	label: Optional[str]
	uri: Optional[URI]


@dataclass(frozen=True)
class LatLonBox:
	min: Position
	max: Position
	label: Optional[str]
	uri: Optional[URI]


@dataclass(frozen=True)
class GeoTrack:
	points: list[Position]
	label: Optional[str]
	uri: Optional[URI]


@dataclass(frozen=True)
class Polygon:
	vertices: list[Position]
	label: Optional[str]
	uri: Optional[URI]


@dataclass(frozen=True)
class Circle:
	center: Position
	radius: float
	label: Optional[str]
	uri: Optional[URI]


@dataclass(frozen=True)
class Pin:
	position: Position
	kind: PinKind


GeoFeature: TypeAlias = FeatureCollection | Position | LatLonBox | GeoTrack | Polygon | Circle | Pin

@dataclass(frozen=True)
class TimeInterval:
	start: Instant
	stop: Instant


@dataclass(frozen=True)
class TemporalCoverage:
	interval: TimeInterval
	resolution: Optional[str]


@dataclass(frozen=True)
class PlainStaticCollection:
	res: URI
	hash: Sha256Sum
	title: str


@dataclass(frozen=True)
class PlainStaticObject:
	res: URI
	hash: Sha256Sum
	name: str


@dataclass(frozen=True)
class StaticCollection:
	res: URI
	hash: Sha256Sum
	members: list[PlainStaticItem]
	creator: Organization
	title: str
	description: Optional[str]
	previousVersion: Optional[URI]
	nextVersion: Optional[URI | list[URI]]
	latestVersion: URI | list[URI]
	parentCollections: list[UriResource]
	doi: Optional[str]
	coverage: Optional[GeoFeature]
	documentation: Optional[PlainStaticObject]
	references: References


PlainStaticItem: TypeAlias = PlainStaticCollection | PlainStaticObject

DataItemCollection: TypeAlias = StaticCollection

DataItem: TypeAlias = PlainStaticItem | DataItemCollection

@dataclass(frozen=True)
class UriResource:
	uri: URI
	label: Optional[str]
	comments: list[str]


@dataclass(frozen=True)
class LinkBox:
	name: str
	coverImage: URI
	target: URI
	orderWeight: Optional[int]


@dataclass(frozen=True)
class WebpageElements:
	self: UriResource
	coverImage: Optional[URI]
	linkBoxes: Optional[list[LinkBox]]


@dataclass(frozen=True)
class Organization:
	self: UriResource
	name: str
	email: Optional[str]
	website: Optional[URI]
	webpageDetails: Optional[WebpageElements]


@dataclass(frozen=True)
class Person:
	self: UriResource
	firstName: str
	lastName: str
	email: Optional[str]
	orcid: Optional[Orcid]


@dataclass(frozen=True)
class Site:
	self: UriResource
	ecosystem: UriResource
	location: Optional[GeoFeature]


@dataclass(frozen=True)
class Project:
	self: UriResource
	keywords: Optional[list[str]]


@dataclass(frozen=True)
class DataTheme:
	self: UriResource
	icon: URI
	markerIcon: Optional[URI]


@dataclass(frozen=True)
class ObjectFormat:
	self: UriResource
	goodFlagValues: Optional[list[str]]


@dataclass(frozen=True)
class DataObjectSpec:
	self: UriResource
	project: Project
	theme: DataTheme
	format: ObjectFormat
	encoding: UriResource
	dataLevel: int
	specificDatasetType: DatasetType
	datasetSpec: Optional[DatasetSpec]
	documentation: list[PlainStaticObject]
	keywords: Optional[list[str]]


@dataclass(frozen=True)
class DatasetSpec:
	self: UriResource
	resolution: Optional[str]


@dataclass(frozen=True)
class DataAcquisition:
	station: Station
	site: Optional[Site]
	interval: Optional[TimeInterval]
	instrument: Optional[UriResource | list[UriResource]]
	samplingPoint: Optional[Position]
	samplingHeight: Optional[float]


@dataclass(frozen=True)
class DataProduction:
	creator: Agent
	contributors: list[Agent]
	host: Optional[Organization]
	comment: Optional[str]
	sources: list[PlainStaticObject]
	documentation: Optional[PlainStaticObject]
	dateTime: Instant


@dataclass(frozen=True)
class DataSubmission:
	submitter: Organization
	start: Instant
	stop: Optional[Instant]


@dataclass(frozen=True)
class StationTimeSeriesMeta:
	acquisition: DataAcquisition
	productionInfo: Optional[DataProduction]
	nRows: Optional[int]
	coverage: Optional[GeoFeature]
	columns: Optional[list[VarMeta]]


@dataclass(frozen=True)
class ValueType:
	self: UriResource
	quantityKind: Optional[UriResource]
	unit: Optional[str]


@dataclass(frozen=True)
class VarMeta:
	model: UriResource
	label: str
	valueType: ValueType
	valueFormat: Optional[URI]
	isFlagFor: Optional[list[URI]]
	minMax: Optional[tuple[float, float]]
	instrumentDeployments: Optional[list[InstrumentDeployment]]


@dataclass(frozen=True)
class SpatioTemporalMeta:
	title: str
	description: Optional[str]
	spatial: GeoFeature
	temporal: TemporalCoverage
	station: Optional[Station]
	samplingHeight: Optional[float]
	productionInfo: DataProduction
	variables: Optional[list[VarMeta]]


@dataclass(frozen=True)
class DataObject:
	hash: Sha256Sum
	accessUrl: Optional[URI]
	pid: Optional[str]
	doi: Optional[str]
	fileName: str
	size: Optional[int]
	submission: DataSubmission
	specification: DataObjectSpec
	specificInfo: SpatioTemporalMeta | StationTimeSeriesMeta
	previousVersion: Optional[URI | list[URI]]
	nextVersion: Optional[URI | list[URI]]
	latestVersion: URI | list[URI]
	parentCollections: list[UriResource]
	references: References


@dataclass(frozen=True)
class DocObject:
	hash: Sha256Sum
	accessUrl: Optional[URI]
	pid: Optional[str]
	doi: Optional[str]
	fileName: str
	size: Optional[int]
	description: Optional[str]
	submission: DataSubmission
	previousVersion: Optional[URI | list[URI]]
	nextVersion: Optional[URI | list[URI]]
	latestVersion: URI | list[URI]
	parentCollections: list[UriResource]
	references: References


@dataclass(frozen=True)
class Licence:
	url: URI
	name: str
	webpage: URI
	baseLicence: Optional[URI]


@dataclass(frozen=True)
class References:
	citationString: Optional[str]
	citationBibTex: Optional[str]
	citationRis: Optional[str]
	doi: Optional[DoiMeta]
	keywords: Optional[list[str]]
	authors: Optional[list[Agent]]
	title: Optional[str]
	temporalCoverageDisplay: Optional[str]
	acknowledgements: Optional[list[str]]
	licence: Optional[Licence]


Agent: TypeAlias = Organization | Person

StaticObject: TypeAlias = DataObject | DocObject

FunderIdType: TypeAlias = Literal["Crossref Funder ID" , "GRID" , "ISNI" , "ROR" , "Other"]

IcosStationClass: TypeAlias = Literal["1" , "2" , "Associated"]

@dataclass(frozen=True)
class Station:
	org: Organization
	id: str
	location: Optional[Position]
	coverage: Optional[GeoFeature]
	responsibleOrganization: Optional[Organization]
	pictures: list[URI]
	specificInfo: StationSpecifics
	countryCode: Optional[CountryCode]
	funding: Optional[list[Funding]]


@dataclass(frozen=True)
class Funding:
	self: UriResource
	funder: Funder
	awardTitle: Optional[str]
	awardNumber: Optional[str]
	awardUrl: Optional[URI]
	start: Optional[LocalDate]
	stop: Optional[LocalDate]


@dataclass(frozen=True)
class Funder:
	org: Organization
	id: Optional[tuple[str, FunderIdType]]


NoStationSpecifics: TypeAlias = None

@dataclass(frozen=True)
class SitesStationSpecifics:
	sites: list[Site]
	ecosystems: list[UriResource]
	climateZone: Optional[UriResource]
	meanAnnualTemp: Optional[float]
	meanAnnualPrecip: Optional[float]
	operationalPeriod: Optional[str]
	discontinued: bool
	documentation: list[PlainStaticObject]


@dataclass(frozen=True)
class AtcStationSpecifics:
	wigosId: Optional[str]
	theme: Optional[DataTheme]
	stationClass: Optional[IcosStationClass]
	labelingDate: Optional[LocalDate]
	discontinued: bool
	timeZoneOffset: Optional[int]
	documentation: list[PlainStaticObject]


@dataclass(frozen=True)
class OtcStationSpecifics:
	theme: Optional[DataTheme]
	stationClass: Optional[IcosStationClass]
	labelingDate: Optional[LocalDate]
	discontinued: bool
	timeZoneOffset: Optional[int]
	documentation: list[PlainStaticObject]


@dataclass(frozen=True)
class EtcStationSpecifics:
	theme: Optional[DataTheme]
	stationClass: Optional[IcosStationClass]
	labelingDate: Optional[LocalDate]
	discontinued: bool
	climateZone: Optional[UriResource]
	ecosystemType: Optional[UriResource]
	meanAnnualTemp: Optional[float]
	meanAnnualPrecip: Optional[float]
	meanAnnualRad: Optional[float]
	stationDocs: list[URI]
	stationPubs: list[URI]
	timeZoneOffset: Optional[int]
	documentation: list[PlainStaticObject]


@dataclass(frozen=True)
class IcosCitiesStationSpecifics:
	timeZoneOffset: Optional[int]


EcoStationSpecifics: TypeAlias = SitesStationSpecifics | EtcStationSpecifics

IcosStationSpecifics: TypeAlias = AtcStationSpecifics | OtcStationSpecifics | EtcStationSpecifics

StationSpecifics: TypeAlias = NoStationSpecifics | EcoStationSpecifics | IcosStationSpecifics | IcosCitiesStationSpecifics

@dataclass(frozen=True)
class InstrumentDeployment:
	instrument: UriResource
	station: Organization
	pos: Optional[Position]
	variableName: Optional[str]
	forProperty: Optional[UriResource]
	start: Optional[Instant]
	stop: Optional[Instant]


@dataclass(frozen=True)
class Instrument:
	self: UriResource
	model: str
	serialNumber: str
	name: Optional[str]
	vendor: Optional[Organization]
	owner: Optional[Organization]
	parts: list[UriResource]
	partOf: Optional[UriResource]
	deployments: list[InstrumentDeployment]


DatasetType: TypeAlias = Literal["StationTimeSeries" , "SpatioTemporal"]


CPJson = TypeVar('CPJson')

def parse_cp_json(input_text: str, data_class: Type[CPJson]) -> CPJson:
	def tuple_hook(d: dict[str, Any]) -> Any:
		for k in d.keys():
			if isinstance(d[k], list):
					d[k] = tuple(d[k])

		return d

	input_dict=json.JSONDecoder(object_hook=tuple_hook).decode(input_text)

	return from_dict(data_class=data_class, data=input_dict, config=Config(cast=[list]))



from dataclasses import dataclass
from typing import Optional
from .metacore import UriResource, Person, Organization, Station

@dataclass(frozen=True)
class RoleDetails:
	role: UriResource
	start: Optional[str]
	end: Optional[str]
	weight: Optional[int]
	extra: Optional[str]

@dataclass(frozen=True)
class PersonRole:
	org: UriResource
	role: RoleDetails

@dataclass(frozen=True)
class PersonWithRoles(Person):
	roles: list[PersonRole]

@dataclass(frozen=True)
class Membership:
	person: Person
	role: RoleDetails

@dataclass(frozen=True)
class OrgWithStaff(Organization):
	staff: list[Membership]

@dataclass(frozen=True)
class StationWithStaff(Station):
	staff: list[Membership]



import json
from urllib import request
from urllib.request import Request, HTTPError, urlopen
from urllib.parse import urlencode
from http.client import HTTPResponse, HTTPMessage
from typing import Any, Literal, IO
from .auth import AuthTokenProvider

Method = Literal["GET", "POST"]

def http_request(
	url: str,
	error_hint: str,
	method: Method = "GET",
	headers: dict[str, str] = {},
	data: str | dict[str, Any] | None = None
) -> HTTPResponse:

	h = {k.lower(): v for k, v in headers.items()}
	cont_type_h = "content-type"

	class NoRedirect(request.HTTPRedirectHandler):
		def redirect_request(self, req: Request, fp: IO[bytes], code: int, msg: str, headers: HTTPMessage, newurl: str) -> Request | None:
			return None

	opener = request.build_opener(NoRedirect)
	request.install_opener(opener)

	if isinstance(data, str):
		if not cont_type_h in h:
			h[cont_type_h] = "text/plain"
		d = bytes(data, encoding="utf-8")
	elif isinstance(data, dict):
		if h.get(cont_type_h) == "application/x-www-form-urlencoded":
			d = bytes(urlencode(data), encoding="utf-8")
		else:
			h[cont_type_h] = "application/json"
			d = json.dumps(data).encode("utf-8")
	else:
		d = None

	req = Request(url=url, data=d, headers=h, method=method)

	try:
		return urlopen(req)
	except HTTPError as err:
		resp_text = err.read().decode()
		if err.status is not None and err.status >= 300 and err.status < 400 and "licence" in resp_text.lower():
			resp_text = "Accepting data licence in your user profile is required for data downloads"
		msg = f"{error_hint} at {url} problem, HTTP response code: {err.status}, response: {resp_text}"
		raise Exception(msg) from None

def http_auth_request(
	url: str,
	error_hint: str,
	auth: AuthTokenProvider,
	method: Method = "GET",
	headers: dict[str, str] = {},
	data: str | dict[str, Any] | None = None
) -> HTTPResponse:
	h = headers.copy()
	h['Cookie'] = auth.get_token().cookie_value
	return http_request(url, error_hint, method, h, data)



from dataclasses import dataclass
from ..sparql import Binding, as_uri, as_string, as_opt_str, as_opt_double, as_opt_float
from ..envri import EnvriConfig
from ..metacore import UriResource

@dataclass(frozen=True)
class StationLite(UriResource):
	"""
	Dataclass for basic metadata of a measurement station.

	Attributes:
		`uri` (str): URI id of the station
		`id` (str): station's id within the measurement station network
		`type_uri` (str): direct type of the station (rdf:type in RDF
			terms)
		`name` (str): station name
		`country_code` (str): ISO 3166-1 alpha-2
		`lat` (float | None): optional WGS-84 latitude of the principal
			station location
		`lon` (float | None): optional WGS-84 longitude of the principal
			station location
		`elevation` (float | None): elevation above sea level in meters
		`geo_json` (str | None): optional string with GeoJSON
			representing station coverage
	"""
	uri: str
	id: str
	type_uri: str
	name: str
	country_code: str
	lat: float | None
	lon: float | None
	elevation: float | None
	geo_json: str | None

def parse_station(row: Binding) -> StationLite:
	station_name = as_string("stationName", row)
	station_id = as_string("stId", row)

	return StationLite(
		uri = as_uri("station", row),
		id = station_id,
		type_uri = as_uri("stType", row),
		name = station_name,
		country_code = as_string("cCode", row),
		lat = as_opt_double("lat", row),
		lon = as_opt_double("lon", row),
		elevation = as_opt_float("elevation", row),
		geo_json = as_opt_str("geoJson", row),
		label = f"{station_name} ({station_id})",
		comments = [],
	)

def station_lite_list(station_type_uri: str | None, conf: EnvriConfig) -> str:
	top_station_class = '<' + station_type_uri + '>' if station_type_uri else 'cpmeta:Station'
	return f"""
		PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
		PREFIX cpmeta: <http://meta.icos-cp.eu/ontologies/cpmeta/>
		SELECT * WHERE{{
			?stType rdfs:subClassOf* {top_station_class} . 
			?station a ?stType ; cpmeta:hasStationId ?stId ; cpmeta:hasName ?stationName ; cpmeta:countryCode ?cCode .
			FILTER(strstarts(str(?station), "{conf.meta_instance_prefix}"))
			OPTIONAL{{?station cpmeta:hasLatitude ?lat ; cpmeta:hasLongitude ?lon}}
			OPTIONAL{{?station cpmeta:hasElevation ?elevation}}
			OPTIONAL{{?station cpmeta:hasSpatialCoverage/cpmeta:asGeoJSON ?geoJson}}
		}}
		ORDER BY ?stId
		"""



from .envri import EnvriConfig
from .sparql import SparqlResults, sparql_select as sparql_select_generic
from .queries.speclist import dobj_spec_lite_list, parse_dobj_spec_lite, DobjSpecLite
from .queries.dataobjlist import DataObjectLite, parse_dobj_lite, dataobj_lite_list
from .queries.dataobjlist import Filter, OrderBy, OrderByProp, CategorySelector
from .queries.stationlist import station_lite_list, parse_station, StationLite
from .metacore import CPJson, parse_cp_json
from .metacore import DataObject as VanillaDataObject, StaticCollection as VanillaStaticCollection
from .rolemeta import StationWithStaff
from .geofeaturemeta import GeoFeatureWithGeo, Point
from .http import http_request
from typing import Type, TypeAlias, Literal, Any, Optional
from dataclasses import dataclass

# the following are needed for convenient re-export by the end user
from .queries.dataobjlist import TimeFilter, SizeFilter, SamplingHeightFilter, GeoIntersectFilter
from .queries.dataobjlist import box_intersect
# and the following to suppress warnings about unused imports:
_needed_classes: TypeAlias = TimeFilter | SizeFilter | SamplingHeightFilter | GeoIntersectFilter | Point
_needed_method = box_intersect

@dataclass(frozen=True)
class DataObject(VanillaDataObject):
	coverageGeo: dict[str, Any]

@dataclass(frozen=True)
class StaticCollection(VanillaStaticCollection):
	coverage: Optional[GeoFeatureWithGeo]

@dataclass(frozen=True)
class Station(StationWithStaff):
	coverage: Optional[GeoFeatureWithGeo]

class MetadataClient:
	def __init__(self, envri_conf: EnvriConfig):
		self._envri_conf = envri_conf

	def sparql_select(self, query: str, disable_cache: bool = False) -> SparqlResults:
		"""
		Runs a SPARQL SELECT query against the SPARQL endpoint exposed
		by the Repository.

		:param `query`:
			A string containing the query
		:param `disable_cache`:
			A boolean flag indicating whether to accept cached SPARQL
			responses or disable the cache. By default the cache is
			enabled.
		:return:
			An instance of `SparqlResults` data class, declared in
			`sparql` module of this library.
		"""
		endpoint = self._envri_conf.sparql_endpoint
		return sparql_select_generic(endpoint, query, disable_cache)

	def list_datatypes(self) -> list[DobjSpecLite]:
		"""
		List available data types. Presense of data of these types is
		not guaranteed.

		:return:
			List of `DobjSpecLite` objects with basic information about
			the data types: URI id `uri`, `label`, `project`, `theme`,
			`data_level`, `dataset_type` (`'StationTimeSeries'` for
			station-specific time series dataset or `'SpatioTemporal'`
			for a more general dataset with spatial and temporal
			coverages, which can optionally be associated with a
			station), and optional dataset specification URI
			`dataset_spec_uri`, presense of which indicates that data
			of this type must have certain column/variable metadata.
			Data objects with `'StationTimeSeries'` datasets and non-
			empty `dataset_spec_uri` can be fetched as CSV or as a
			dictionary of readily parsed arrays using DataClient
			(obtainable as e.g. `from icoscp_core.<repo> import data`).
		"""
		query = dobj_spec_lite_list(self._envri_conf)
		qres = self.sparql_select(query)
		return [parse_dobj_spec_lite(b) for b in qres.bindings]

	def list_stations(self, of_station_type_uri: str | None | Literal[False] = None ) -> list[StationLite]:
		"""
		List basic information about available stations. Presense of
		data from/for these stations is not guaranteed, as there may
		not have been any data from this station uploaded.

		:param `of_station_type_uri`: The URI of the type of
		interesting station (in RDF metadata, the rdf:type of the
		station resource, or a supertype of the interesting types).
		If omitted, a default type specific to the Repository
		context is used (e.g. only the ICOS stations for ICOS
		Repository). Look in `icoscp_core.<repo>` package for your
		Repository for defined constants for different interesting
		types of stations. For example, `icoscp_core.icos.ATMO_STATION`
		for ICOS Atmosphere stations. To get all stations without any
		filtering, use `False`.

		:return: A list of `StationLite` objects with basic
		information about the stations, defined in package
		`queries.stationlist`
		"""

		station_type: str | None = None
		if of_station_type_uri is None:
			station_type = self._envri_conf.default_station_type_url
		elif type(of_station_type_uri) == str:
			station_type = of_station_type_uri

		query = station_lite_list(station_type, self._envri_conf)
		qres = self.sparql_select(query)
		return [parse_station(b) for b in qres.bindings]

	def list_data_objects(
		self,
		datatype: CategorySelector = None,
		station: CategorySelector = None,
		filters: list[Filter] = [],
		include_deprecated: bool = False,
		order_by: OrderBy | OrderByProp | None = {"prop": "submTime", "descending": True},
		limit: int = 100,
		offset: int = 0
	) -> list[DataObjectLite]:
		"""
		Method for listing, selecting, filtering, sorting, and paging
		through the lists of, data objects.

		:param `datatype`: either a `DobjSpecLite` object (obtainable
		from `list_datatypes` method), or a URI identifying a datatype,
		or a list of either of the two, or `None` (default) for all
		datatypes.

		:param `station`: either a `StationLite` object (obtainable
		from `list_stations` method), or a URI identifying a station, or
		a list of either of the two, or `None` (default) for all
		stations.

		:param `filters`: a list of filters, each filter being either
		`TimeFilter` for filtering by submission time, or data start- or
		stop time, `SizeFilter` for filtering by file size,
		`SamplingHeightFilter` for filtering by sampling height
		(applicable only to data objects with sampling height metadata),
		or `GeoIntersectFilter` for geospatial filtering;
		the list is empty by default.

		:param `include_deprecated`: boolean flag to include deprecated
		(i.e. such that have a newer version with explicit metadata
		link) data objects in the list; `False` by default.

		:param `order_by`: either `OrderByProp` --- a property to sort
		the objects by, in ascending order (one of strings 'submTime',
		'timeStart', 'timeEnd', 'size', 'fileName') or a dictionary of
		the form {"prop": OrderByProp, "descending": bool}, or None for
		no sorting. By default, the results are sorted by submission
		time in descending order.

		:param `limit`: the maximum size of the list to return; 100 by
		default; values above 10000 are ignored and 10000 is used.

		:param `offset`: the results page offset; 0 by default.

		:return: a list of `DataObjectLite` instances with basic data
		object metadata (defined in `queries.dataobjlist` package).
		"""
		query = dataobj_lite_list(datatype, station, filters, include_deprecated, order_by, limit, offset)
		qres = self.sparql_select(query)
		return [parse_dobj_lite(b) for b in qres.bindings]

	def get_dobj_meta(self, dobj: str | DataObjectLite) -> DataObject:
		"""
		Get fully detailed metadata of a data object

		:param `dobj`: either a landing page URI or an instance of
		`DataObjectLite` class (obtainable with `list_data_objects`
		method)

		:return: Python dataclass `DataObject` whose definition was
		generated by automatically transforming the corresponding data
		portal code; it closely mimicks the JSON metadata returned by
		the data portal back end from the data object landing page.
		"""
		dobj_uri: str

		if type(dobj) == str: dobj_uri = dobj
		elif type(dobj) == DataObjectLite: dobj_uri = dobj.uri
		else: raise ValueError("dobj must be either landing page URL or an instance of DataObjectLite")

		return _get_json_meta(dobj_uri, DataObject)
	
	def get_station_meta(self, station: str | StationLite) -> Station:
		"""
		Get fully detailed metadata of a station

		:param `station`: either a station URI id or an instance of
		`StationLite` class (obtainable with `list_stations` method)

		:return: Python dataclass `Station` whose definition was
		generated by automatically transforming the corresponding data
		portal code; it closely mimicks the JSON metadata returned by
		the data portal back end from the station URI id.
		"""
		station_uri: str

		if type(station) == str: station_uri = station
		elif type(station) == StationLite: station_uri = station.uri
		else: raise ValueError("Station must be either landing page URL or an instance of StationLite")

		return _get_json_meta(station_uri, Station)

	def get_collection_meta(self, coll_uri: str) -> StaticCollection:
		"""
		Get metadata of a collection

		:param `coll_uri`: a string with URI of the collection's landing page

		:return: Python dataclass `StaticCollection` whose definition was
		generated by automatically transforming the corresponding data
		portal code; it closely mimicks the JSON metadata returned by
		the data portal back end from the collection URI id.
		"""
		return _get_json_meta(coll_uri, StaticCollection)

def _get_json_meta(url: str, data_class: Type[CPJson]) -> CPJson:
	headers = {"Accept": "application/json"}
	resp = http_request(url, "Fetching JSON metadata", headers=headers)
	return parse_cp_json(resp.read().decode(), data_class=data_class)



import io
import numpy as np
from numpy.typing import NDArray
import os
import re
import sys
from typing import Any, TypeAlias, TypedDict, Tuple
from dataclasses import dataclass
from .metaclient import MetadataClient
from .metacore import DataObject, StationTimeSeriesMeta, URI
from .queries.dataobjlist import DataObjectLite
from .queries.cpbmeta import CpbMetaData, DatasetCol, get_cpb_meta, get_dataset_cols, get_good_flags_per_spec

@dataclass(frozen=True)
class ColumnInfo:
	label: str
	value_format_uri: URI
	flag_col: str | None = None

@dataclass(frozen=True)
class TableRequest:
	desired_columns: list[str] | None
	offset: int | None
	length: int | None

@dataclass(frozen=True)
class CodecInfo(TableRequest):
	dobj_hash_id: str
	obj_format_uri: URI
	columns: list[ColumnInfo]
	n_rows: int
	good_flags: list[str] | None

def codec_from_dobj_meta(dobj: DataObject, request: TableRequest) -> "Codec":
	spec_info = dobj.specificInfo
	if not isinstance(spec_info, StationTimeSeriesMeta):
			raise Exception(f"Object {dobj.hash} is not a station-specific time series data object.")
	cols_meta = spec_info.columns
	if cols_meta is None:
		raise Exception(f"Metadata of object '{dobj.hash}' does not include any column information")
	col_flags: dict[URI, str] = {
		flagged_col: col.label
		for col in cols_meta if col.isFlagFor is not None
		for flagged_col in col.isFlagFor
	}
	cols_meta_parsed = [
		ColumnInfo(
			label=col.label,
			value_format_uri=col.valueFormat,
			flag_col=col_flags.get(col.model.uri)
		)
		for col in cols_meta
		if col.valueFormat is not None
	]
	if len(cols_meta_parsed) == 0:
		raise Exception(f"Metadata of object '{dobj.hash}' does not include any column for which value format is provided")
	n_rows = spec_info.nRows
	if n_rows is None:
		raise Exception(f"Metadata of object '{dobj.hash}' does not include the number of table rows")
	ci = CodecInfo(
		dobj_hash_id=dobj.hash[:24],
		obj_format_uri=dobj.specification.format.self.uri,
		columns=cols_meta_parsed,
		n_rows=n_rows,
		good_flags=dobj.specification.format.goodFlagValues,
		desired_columns=request.desired_columns,
		offset=request.offset,
		length=request.length
	)
	return Codec(ci)

Dobj: TypeAlias = URI | DataObjectLite

def to_dobj_uri(dobj: Dobj) -> URI:
	if isinstance(dobj, DataObjectLite): return dobj.uri
	elif isinstance(dobj, URI): return dobj
	else: raise ValueError('dobj must be a string or a DataObjectLite instance')

def codecs_from_dobjs(dobjs: list[Dobj], request: TableRequest, meta: MetadataClient) -> list[Tuple[Dobj, "Codec"]]:

	if len(dobjs) == 0: raise Exception("Got an empty list of data objects")

	dobj_uris = [to_dobj_uri(dobj) for dobj in dobjs]

	cpb_metas: dict[URI, CpbMetaData] = {
		cpb.dobj: cpb for cpb in get_cpb_meta(dobj_uris, meta)
	}

	missing_dobjs = [str(dobj) for dobj in dobjs if to_dobj_uri(dobj) not in cpb_metas]

	if len(missing_dobjs) > 0: raise Exception(
		"Some of the requested objects don't have required metadata to fetch them as tabular data: " +
		",\n".join(missing_dobjs)
	)

	dataset_specs = {cpb.dataset_spec for cpb in cpb_metas.values()}

	if len(dataset_specs) > 1: raise Exception(
		"When requesting data from several data objects, all data objects "
		"must have a common dataset specification"
	)
	dataset_spec = dataset_specs.pop()

	dataset_cols = get_dataset_cols(dataset_spec, meta)
	col_lookup = ColumnLookupByLabel(dataset_cols, dataset_spec)

	dobj_specs = [cpb.obj_spec for cpb in cpb_metas.values()]
	good_flags_lookup: dict[URI, list[str]] = get_good_flags_per_spec(dobj_specs, meta)

	res: list[Tuple[Dobj, "Codec"]] = []
	for dobj in dobjs:
		cpb = cpb_metas[to_dobj_uri(dobj)]
		cols_names = cpb.col_names or col_lookup.default_actual_cols

		cols_info = [
			ColumnInfo(lbl, col.val_format, col.flag_col)
			for lbl, col in
				[(lbl, col_lookup.lookup_column(lbl)) for lbl in cols_names]
			if col is not None
		]
		ci = CodecInfo(
			dobj_hash_id=cpb.dobj.split("/")[-1],
			obj_format_uri=cpb.obj_format,
			columns=cols_info,
			good_flags=good_flags_lookup.get(cpb.obj_spec),
			n_rows=cpb.n_rows,
			desired_columns=request.desired_columns,
			offset=request.offset,
			length=request.length
		)
		res.append((dobj, Codec(ci)))
	return res

ArraysDict: TypeAlias = dict[str, NDArray[Any]]

class Codec:
	def __init__(self, ci: CodecInfo) -> None:
		desired_indices: list[int]
		ci.columns.sort(key=lambda col: col.label)
		if ci.desired_columns is None:
			desired_indices = [i for i in range(len(ci.columns))]
		else:
			labels = [col_info.label for col_info in ci.columns]
			def get_col_idx(col_lbl: str) -> int:
				try:
					return labels.index(col_lbl)
				except ValueError:
					raise ValueError(f'Column "{col_lbl}" is not actually present in data object {ci.dobj_hash_id}')
			expl_desired_indices = {get_col_idx(desired_col) for desired_col in ci.desired_columns}
			flag_indices = {
				get_col_idx(flag_col)
				for flag_col in [ci.columns[i].flag_col for i in expl_desired_indices]
				if flag_col is not None
			}
			desired_indices = [i for i in expl_desired_indices | flag_indices]
			desired_indices.sort()

		# The following checks are needed because the backend does not return user-friendly error messages or results.
		for param, val in [("offset", ci.offset), ("length", ci.length)]:
			if val is not None and val > ci.n_rows:
				raise ValueError(f"The value provided for the '{param}' parameter ({val}) is larger than the number of available rows ({ci.n_rows})")
		if ci.offset is not None and ci.length is not None and ci.n_rows - ci.offset < ci.length:
			raise ValueError(
				f"The value provided for the 'length' parameter ({ci.length}) is larger than the number of available rows ({ci.n_rows}) "
				f"minus the value provided for the 'offset' parameter ({ci.offset})"
			)

		json: dict[str, Any] = {
			"tableId": ci.dobj_hash_id,
			"subFolder": _subfolder_path(ci),
			"schema": {
				"columns": [_get_fmt(col_info.value_format_uri) for col_info in ci.columns],
				"size": ci.n_rows
			},
			"columnNumbers": desired_indices
		}
		if ci.offset is not None or ci.length is not None:
			offset = ci.offset or 0
			length = ci.length or (ci.n_rows - offset)
			json["slice"] = {
				"offset": offset,
				"length": length
			}
		self._ci = ci
		self._json_payload = json
		self._desired_indices = desired_indices

	@property
	def json_payload(self):
		return self._json_payload

	def parse_cpb_response(self, resp: io.BufferedReader, keep_bad_data: bool) -> ArraysDict:
		return self._parse_from_buff(resp, None, keep_bad_data)

	def parse_cpb_file(self, data_folder_path: str, keep_bad_data: bool) -> ArraysDict:

		file_path = os.path.join(
			data_folder_path,
			_subfolder_path(self._ci),
			self._ci.dobj_hash_id + ".cpb"
		)

		col_offsets = _column_offsets(self._ci)

		with open(file_path, "rb") as file:
			return self._parse_from_buff(file, col_offsets, keep_bad_data)


	def _parse_from_buff(self, buff: io.BufferedReader, col_offsets: list[int] | None, keep_bad_data: bool) -> ArraysDict:
		res: ArraysDict = {}
		cols = self._ci.columns
		n_fetch = _n_rows_to_fetch(self._ci)
		for n in self._desired_indices:
			col = cols[n]
			fmt = _get_fmt(col.value_format_uri)
			must_read = _get_byte_size(fmt) * n_fetch
			if col_offsets is not None:
				buff.seek(col_offsets[n])
			col_bytes = buff.read(must_read)
			if must_read > len(col_bytes):
				raise IOError(f'Error while reading cpb response for column {col.label} ({n_fetch} values), reached end of data, but got only {len(col_bytes)} bytes instead of {must_read}')
			arr = np.frombuffer(col_bytes, dtype = _get_format_dtype(fmt))
			arr = _type_post_process(arr, col.value_format_uri)

			if sys.byteorder == 'little':
				# cpb files are all big-endian, fixing for Pandas compatibility on little endian systems:
				arr = arr.byteswap().newbyteorder()
			res[col.label] = arr

		if not keep_bad_data and self._ci.good_flags:
			flag_masks: dict[str, NDArray[Any]] = {}
			desired_cols = [cols[n] for n in self._desired_indices]
			flag_cols = [
				flag for flag in { # to make the list unique
					col.flag_col
					for col in desired_cols
					if col.flag_col
				}
			]
			good_flags = self._ci.good_flags
			for flag in flag_cols:
				flag_arr = res[flag]
				submasks = [flag_arr == good_flag for good_flag in good_flags]
				flag_masks[flag] = ~np.logical_or.reduce(submasks)
			for col in desired_cols:
				if col.flag_col:
					arr = res[col.label]
					arr[flag_masks[col.flag_col]] = np.NAN
		return res


def _column_offsets(ci: CodecInfo) -> list[int]:
	col_cell_sizes: list[int] = [
		_get_byte_size(_get_fmt(col.value_format_uri))
		for col in ci.columns
	]
	col_offset: int = 0
	col_offsets: list[int] = []
	n_skip = ci.offset or 0
	for cell_size in col_cell_sizes:
		col_offsets.append(col_offset + n_skip * cell_size)
		col_offset += cell_size * ci.n_rows
	return col_offsets


def _n_rows_to_fetch(ci: CodecInfo) -> int:
	n_rows = ci.n_rows - (ci.offset or 0)
	if ci.length: n_rows = min(n_rows, ci.length)
	return n_rows


def _subfolder_path(ci: CodecInfo) -> str:
	return ci.obj_format_uri.split("/")[-1]

class RegexCol(TypedDict):
	regex: re.Pattern[str]
	col: DatasetCol

class ColumnLookupByLabel:
	def __init__(self, ds_cols: list[DatasetCol], ds_spec_uri: URI) -> None:
		if len(ds_cols) == 0:
			raise Exception(f"Dataset specification '{ds_spec_uri}' has no columns")
		self._ds_spec_uri = ds_spec_uri
		self._default_actual_cols: list[str] = [col.col_title for col in ds_cols if not (col.is_optional or col.is_regex)]
		self._verbatim_lookup: dict[str, DatasetCol] = {col.col_title: col for col in ds_cols if not col.is_regex}
		self._regex_cols: list[RegexCol] = [
			{"regex": re.compile(col.col_title), "col": col}
			for col in ds_cols if col.is_regex
		]

	@property
	def default_actual_cols(self) -> list[str]:
		res = self._default_actual_cols
		if len(res) == 0:
			raise Exception(f"No mandatory verbatim columns in dataset spec '{self._ds_spec_uri}'")
		return res

	def lookup_column(self, col_label: str) -> DatasetCol | None:
		val_format = self._verbatim_lookup.get(col_label)
		if val_format: return val_format
		for re_col in self._regex_cols:
			if re_col["regex"].match(col_label):
				return re_col["col"]

_fmt_uri_to_fmt = {
	'int32':'INT',
	'float32': 'FLOAT',
	'float64': 'DOUBLE',
	'bmpChar': 'CHAR',
	'string': 'STRING',
	'iso8601date': 'INT',
	'etcDate': 'INT',
	'iso8601month': 'INT',
	'iso8601timeOfDay': 'INT',
	'iso8601dateTime': 'DOUBLE',
	'isoLikeLocalDateTime': 'DOUBLE',
	'etcLocalDateTime': 'DOUBLE'
}

def _get_fmt(fmt_uri: str) -> str:
	return _fmt_uri_to_fmt[fmt_uri.split("/")[-1]]

def _type_post_process(arr: NDArray[Any], fmt: URI) -> NDArray[Any]:
	fmt_str = fmt.split("/")[-1]
	if fmt_str == "bmpChar":
		return arr.astype('>u4').view(dtype='>U1')
	elif fmt_str in ["iso8601date", "etcDate"]:
		return arr.astype('>i8').view(dtype=np.dtype('>datetime64[D]'))
	elif fmt_str == "iso8601month":
		return (arr - 1970 * 12).astype('>i8').view(dtype=np.dtype('>datetime64[M]'))
	elif fmt_str == "iso8601timeOfDay":
		return arr.astype('>i8').view(dtype=np.dtype('>timedelta64[s]'))
	elif fmt_str in ["iso8601dateTime", "isoLikeLocalDateTime", "etcLocalDateTime"]:
		return arr.astype('>i8').view(dtype=np.dtype('>datetime64[ms]'))
	else:
		return arr


_fmt_to_dtype: dict[str, np.dtype[Any]] = {
	'INT': np.dtype('>i4'),
	'FLOAT': np.dtype('>f4'),
	'DOUBLE': np.dtype('>f8'),
	'SHORT': np.dtype('>i2'),
	'CHAR': np.dtype('>u2'),
	'BYTE': np.dtype('i8'),
}

def _get_format_dtype(fmt: str) -> np.dtype[Any]:
	try:
		return _fmt_to_dtype[fmt]
	except KeyError:
		raise ValueError(f"Unsupported cpb value format {fmt}")

_fmt_to_byte_size = {
	'INT': 4,
	'FLOAT': 4,
	'DOUBLE': 8,
	'SHORT': 2,
	'CHAR': 2,
	'BYTE': 1,
	'STRING': 4
}

def _get_byte_size(fmt: str) -> int:
	return _fmt_to_byte_size[fmt]




from dataclasses import dataclass

from ..sparql import Binding, as_int, as_uri, as_string, as_opt_uri
from ..envri import EnvriConfig
from ..metacore import UriResource, DatasetType


@dataclass(frozen=True)
class DobjSpecLite(UriResource):
	"""
	Dataclass representing  data type (i.e. type of a data object, or data
	object specification). Extends UriResource.

	Attributes:
		`uri` (string):
			URI id of the data type. Inherited from UriResource
		`label` (string):
			user-friendly label of this data type. Inherited from UriResource
		`data_level` (int):
			0 for raw observational data,
			1 for automatically produced observational data products,
			2 for quality-controlled observational data products,
			3 for elaborated products
		`dataset_type` (string):
			'StationTimeSeries' for tabular single-station observational data
			'SpatioTemporal' for objects with both temporal and geo coverage
		`dataset_spec_uri` (string):
			URI reference to metadata about the dataset variables
		`theme` (UriResource):
			repository-specific data theme (e.g. Ocean or Ecosystem for ICOS);
			is itself a dataclass with `uri` and `label` properties
		`project` (UriResource):
			repository-specific data affiliation project (e.g. ICOS, GCP,
			SITES AquaNet);
			is itself a dataclass with `uri` and `label` properties
		`format` (UriResource):
			data object format, e.g. CSV, NetCDF, arbitrary ZIP, etc;
			is itself a dataclass with `uri` and `label` properties
	"""
	data_level: int
	dataset_type: DatasetType
	dataset_spec_uri: str | None
	theme: UriResource
	project: UriResource
	format: UriResource
	@property
	def has_data_access(self) -> bool:
		return self.dataset_type == "StationTimeSeries" and self.dataset_spec_uri is not None

def dobj_spec_lite_list(conf: EnvriConfig) -> str:
	return f"""
prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
prefix cpmeta: <http://meta.icos-cp.eu/ontologies/cpmeta/>
prefix xsd: <http://www.w3.org/2001/XMLSchema#>
select * where{{
	?spec cpmeta:hasDataLevel ?level ;
		cpmeta:hasAssociatedProject ?project ;
		cpmeta:hasFormat ?format ;
		cpmeta:hasSpecificDatasetType ?dsType .
	FILTER NOT EXISTS {{?project cpmeta:hasHideFromSearchPolicy "true"^^xsd:boolean}}
	FILTER(STRSTARTS(str(?spec), "{conf.meta_instance_prefix}"))
	?spec rdfs:label ?specLabel ; cpmeta:hasDataTheme ?theme .
	?theme rdfs:label ?themeLabel .
	?project rdfs:label ?projectLabel .
	?format rdfs:label ?formatLabel .
	optional {{?spec cpmeta:containsDataset ?ds_spec_uri}}
}}
order by ?specLabel"""

def parse_dobj_spec_lite(row: Binding) -> DobjSpecLite:
	dsType: DatasetType
	match as_uri("dsType", row):
		case "http://meta.icos-cp.eu/ontologies/cpmeta/stationTimeSeriesDataset":
			dsType = "StationTimeSeries"
		case "http://meta.icos-cp.eu/ontologies/cpmeta/spatioTemporalDataset":
			dsType = "SpatioTemporal"
		case bad:
			raise ValueError(f"Unexpected URI representing a dataset type: {bad}")

	return DobjSpecLite(
		uri = as_uri("spec", row),
		label = as_string("specLabel", row),
		comments = [],
		data_level = as_int("level", row),
		project = UriResource(
			uri = as_uri("project", row),
			label = as_string("projectLabel", row),
			comments = []
		),
		theme = UriResource(
			uri = as_uri("theme", row),
			label = as_string("themeLabel", row),
			comments = []
		),
		format = UriResource(
			uri = as_uri("format", row),
			label = as_string("formatLabel", row),
			comments = []
		),
		dataset_type = dsType,
		dataset_spec_uri = as_opt_uri("ds_spec_uri", row)
	)



import os
import re
import shutil
from urllib.parse import urlsplit, unquote
from typing import Iterator, Tuple

from .metaclient import MetadataClient
from .metacore import DataObject
from .queries.dataobjlist import DataObjectLite
from .envri import EnvriConfig
from .auth import AuthTokenProvider
from .cpb import Codec, TableRequest, ArraysDict, codec_from_dobj_meta, codecs_from_dobjs
from .cpb import Dobj, to_dobj_uri
from .portaluse_client import report_cpb_file_read
from .http import http_auth_request, HTTPResponse


class DataClient:
	def __init__(
		self,
		conf: EnvriConfig,
		auth: AuthTokenProvider,
		data_folder_path: str | None
	) -> None:
		self._conf = conf
		self._auth = auth
		self._meta = MetadataClient(conf)
		self._data_folder_path = data_folder_path

	@property
	def meta(self) -> MetadataClient:
		return self._meta

	@property
	def auth(self) -> AuthTokenProvider:
		return self._auth

	def get_file_stream(self, dobj: str | DataObjectLite) -> Tuple[str, HTTPResponse]:
		"""
		Fetches the original verbatim content of a data- or document
		object. The method is HTTP-backed and requires authentication
		if fetching data-, not document, objects.

		:param dobj:
			the landing page URI of the object, or a DataObjectLite
			instance
		:returns:
			a tuple of object filename and the HTTPResponse with
			contents, which can be either saved to disk or directly sent
			for processing
		"""

		dobj_uri = to_dobj_uri(dobj)
		url = self._conf.data_service_base_url + urlsplit(dobj_uri).path
		resp = http_auth_request(url, 'Fetching data object', self._auth)

		filename: str
		if type(dobj) is DataObjectLite:
			filename = dobj.filename
		else:
			disp_head = resp.getheader("Content-Disposition") or ""
			fn_match = re.search(r'filename="(.*)"', disp_head)
			if not fn_match:
				raise Exception(f"No filename information in response from {url}")
			filename = unquote(fn_match.group(1))

		return filename, resp

	def save_to_folder(self, dobj_uri: str | DataObjectLite, folder_path: str) -> str:
		"""
		Downloads the original verbatim content of a data- or document
		object, and saves it to a folder in a file named according to
		object's metadata. If the file already exists, it gets
		overwritten.

		Requires authentication for data object downloads, even on a
		Jupyter notebook with data storage folder access.

		:param dobj_uri:
			the landing page URI of the object, or a DataObjectLite
			instance
		:param folder_path:
			path to the target folder

		:returns:
			the name of the file used to save the object
		"""

		filename, resp_buffer = self.get_file_stream(dobj_uri)

		save_path = os.path.join(folder_path, filename)
		with resp_buffer as reader:
			with open(save_path, "wb") as writer:
				shutil.copyfileobj(reader, writer)
				return filename


	def get_csv_byte_stream(
		self,
		dobj: str | DataObjectLite,
		cols: list[str] = [],
		limit: int | None = None,
		offset: int | None = None
	) -> HTTPResponse:
		"""
		Fetches a standardized plain CSV serialization of a tabular
		data object (that has columnar metadata and has been ingested by
		the data portal).

		This method always requires authentication, even if used from a
		Jupyter notebook with access to the data storage folder.

		:param dobj:
			the landing page URI of the data object, or a
			DataObjectLite instance
		:param cols:
			list of columns to be included; if omitted, all known
			columns will be returned
		:param limit:
			limits the number of table rows to be returned
		:param offset:
			number of rows to skip

		:return:
			HTTPResponse with a stream of CSV bytes. Can be readily
			parsed with `pandas.read_csv`
		"""

		dobj_hash = to_dobj_uri(dobj).split('/')[-1]

		return http_auth_request(
			url = self._conf.data_service_base_url + "/csv/" + dobj_hash,
			error_hint = 'Fetching CSV',
			auth=self._auth,
			data={
				"col": cols,
				"offset": offset,
				"limit": limit
			}
		)


	def get_columns_as_arrays(
		self,
		dobj: DataObject,
		columns: list[str] | None = None,
		keep_bad_data: bool = False,
		offset: int | None = None,
		length: int | None = None
	) -> ArraysDict:
		"""
		Fetches a binary tabular data object and returns it as a
		dictionary of numpy arrays with variable names as keys.

		:param dobj:
			a DataObject instance with detailed data object metadata
			(obtainable from `MetaClient`'s method `get_dobj_meta`)
		:param cols:
			list of columns to be included; if None, all known columns
			will be returned; if the requested columns have accompanying
			quality flag columns, the latter will be automatically
			included.
		:param keep_bad_data:
			flag to force including numeric values that have been
			flagged as not good in corresponding quality flag columns;
			by default these values are set to NaN
		:param offset:
			number of heading rows to skip; if None, does not skip any
			row
		:param length:
			number of rows to return; if None, return all rows

		:return:
			a dictionary mapping column names to numpy arrays. The
			dictionary can be readily sent to `pandas.DataFrame`
			constructor.
		"""

		req = TableRequest(desired_columns=columns, offset=offset, length=length)
		codec = codec_from_dobj_meta(dobj, req)
		res = self._get_columns_as_arrays(codec, keep_bad_data)

		if self._data_folder_path is not None:
			report_cpb_file_read(self._conf, self._auth, hashes = [dobj.hash[:24]], columns=columns)

		return res


	def batch_get_columns_as_arrays(
		self,
		dobjs: list[Dobj],
		columns: list[str] | None = None,
		keep_bad_data: bool = False
	) -> Iterator[Tuple[Dobj, ArraysDict]]:
		"""
		Efficient batch-fetching version of `get_columns_as_arrays`
		method. Can only be used with groups of data objects of the same
		data type, or whose data types share a dataset specification.

		:param dobjs:
			either a list of data object landing page URIs, or a list of
			`DataObjectLite` instances (obtainable from `MetaClient`'s
			method `list_data_objects`)

		:param columns:
			a list of names of columns to be fetched, or `None` for all
			preview-available columns in the data objects. If the
			requested columns have accompanying quality flag columns,
			the latter will be automatically included.

		:param keep_bad_data:
			flag to force including numeric values that have been
			flagged as not good in corresponding quality flag columns;
			by default these values are set to NaN

		:return:
			a lazy iterable of pairs of the data objects (echoed back
			from the `dobjs` input) and a dictionary mapping column
			names to numpy arrays.
		"""
		req = TableRequest(columns, None, None)
		dobj_codecs = codecs_from_dobjs(dobjs, req, self._meta)

		if self._data_folder_path is not None:
			hashes = [to_dobj_uri(dobj).split('/')[-1] for dobj, _ in dobj_codecs]
			report_cpb_file_read(self._conf, self._auth, hashes=hashes, columns=columns)

		for dobj, codec in dobj_codecs:
			yield dobj, self._get_columns_as_arrays(codec, keep_bad_data)


	def _get_columns_as_arrays(self, codec: Codec, keep_bad_data: bool) -> ArraysDict:
		if self._data_folder_path is not None:
			return codec.parse_cpb_file(self._data_folder_path, keep_bad_data)

		resp = http_auth_request(
			url=self._conf.data_service_base_url + '/cpb',
			error_hint="Fetching binary",
			auth=self._auth,
			method="POST",
			headers={"Accept": "application/octet-stream"},
			data=codec.json_payload
		)
		return codec.parse_cpb_response(resp.fp, keep_bad_data)



from typing import Literal, TypeAlias
from dataclasses import dataclass

Envri: TypeAlias = Literal["ICOS", "SITES", "ICOSCities"]

@dataclass(frozen=True)
class EnvriConfig:
	envri: Envri
	sparql_endpoint: str
	meta_instance_prefix: str
	auth_service_base_url: str
	data_service_base_url: str
	default_station_type_url: str
	token_name: str
	site_used: bool

ICOS_CONFIG = EnvriConfig(
	envri = "ICOS",
	sparql_endpoint = "https://meta.icos-cp.eu/sparql",
	meta_instance_prefix = "http://meta.icos-cp.eu/",
	auth_service_base_url = "https://cpauth.icos-cp.eu/",
	data_service_base_url = "https://data.icos-cp.eu",
	default_station_type_url = "http://meta.icos-cp.eu/ontologies/cpmeta/IcosStation",
	token_name = "cpauthToken",
	site_used = False
)

SITES_CONFIG = EnvriConfig(
	envri = "SITES",
	sparql_endpoint = "https://meta.fieldsites.se/sparql",
	meta_instance_prefix = "https://meta.fieldsites.se/",
	auth_service_base_url = "https://auth.fieldsites.se/",
	data_service_base_url = "https://data.fieldsites.se",
	default_station_type_url = "https://meta.fieldsites.se/ontologies/sites/Station",
	token_name = "fieldsitesToken",
	site_used = True
)

CITIES_CONFIG = EnvriConfig(
	envri = "ICOSCities",
	sparql_endpoint = "https://citymeta.icos-cp.eu/sparql",
	meta_instance_prefix = "https://citymeta.icos-cp.eu/",
	auth_service_base_url = "https://cpauth.icos-cp.eu/",
	data_service_base_url = "https://citydata.icos-cp.eu",
	default_station_type_url = "http://meta.icos-cp.eu/ontologies/cpmeta/IcosCitiesStation",
	token_name = "cpauthToken",
	site_used = False
)

ENVRIES: dict[Envri, EnvriConfig] = {
	"ICOS": ICOS_CONFIG,
	"SITES": SITES_CONFIG,
	"ICOSCities": CITIES_CONFIG
}



from .envri import CITIES_CONFIG
from .bootstrap import Bootstrap

bootstrap = Bootstrap(CITIES_CONFIG)

auth, meta, data = bootstrap.fromPasswordFile()

MID_LOW_COST_STATION = "http://meta.icos-cp.eu/ontologies/cpmeta/CityMidLowCostStation"
MUNICH_MID_LOW_COST_STATION = "http://meta.icos-cp.eu/ontologies/cpmeta/MunichMidLow"
PARIS_MID_LOW_COST_STATION = "http://meta.icos-cp.eu/ontologies/cpmeta/ParisMidLow"
ZURICH_MID_LOW_COST_STATION = "http://meta.icos-cp.eu/ontologies/cpmeta/ZurichMidLow"



import logging
import threading
from typing import Any
from . import __version__ as icoscp_core_version

from .auth import AuthTokenProvider
from .envri import EnvriConfig
from .http import http_request

def report_cpb_file_read(
	conf: EnvriConfig,
	auth: AuthTokenProvider,
	hashes: list[str],
	columns: list[str] | None
) -> None:
	access_info: dict[str, Any] = {
		"objects": hashes,
		"client": 'icoscp_core',
		"version": icoscp_core_version
	}
	if columns is not None:
		access_info["columns"] = columns

	payload = {"cpbFileAccess": access_info}

	return threading.Thread(target=_report_sync, args=(conf, auth, payload)).start()

def _report_sync(conf: EnvriConfig, auth: AuthTokenProvider, payload: dict[str, Any]) -> None:
	url = conf.auth_service_base_url + 'logs/portaluse'
	headers: dict[str, str] = {}
	try:
		headers['Cookie'] = auth.get_token().cookie_value
	except Exception as err:
		# auth may not be initialized on Jupyter, so do nothing
		pass
	try:
		http_request(url, "Reporting data usage event", "POST", headers, payload)
	except Exception as err:
		logging.warning(str(err))



import json
from dataclasses import dataclass

from ..metaclient import MetadataClient
from ..metacore import URI
from ..sparql import Binding, as_uri, as_int, as_opt_str, as_string, as_opt_bool

@dataclass(frozen=True)
class CpbMetaData:
	dobj: URI
	obj_spec: URI
	dataset_spec: URI
	obj_format: URI
	n_rows: int
	col_names: list[str] | None

@dataclass(frozen=True)
class DatasetCol:
	col_title: str
	val_format: URI
	is_optional: bool
	is_regex: bool
	flag_col: str | None


def get_cpb_meta(dobjs: list[URI], meta: MetadataClient) -> list[CpbMetaData]:
	query = f"""
		prefix cpmeta: <http://meta.icos-cp.eu/ontologies/cpmeta/>
		select * where{{
			values ?dobj {{ <{'> <'.join(dobjs)}> }}
			?dobj cpmeta:hasObjectSpec ?objSpec ; cpmeta:hasNumberOfRows ?nRows .
			?objSpec cpmeta:containsDataset ?dsSpec ; cpmeta:hasFormat ?objFormat .
			optional{{?dobj cpmeta:hasActualColumnNames ?colNames }}
		}}"""
	return [_parse_cpb_meta(binding) for binding in meta.sparql_select(query).bindings]

def _parse_cpb_meta(binding: Binding) -> CpbMetaData:
	col_names_json = as_opt_str("colNames", binding)
	return CpbMetaData(
		dobj=as_uri("dobj", binding),
		obj_spec=as_uri("objSpec", binding),
		dataset_spec=as_uri("dsSpec", binding),
		obj_format=as_uri("objFormat", binding),
		n_rows=as_int("nRows", binding),
		col_names=json.loads(col_names_json) if col_names_json else None
	)

def get_dataset_cols(dataset_spec: URI, meta: MetadataClient) -> list[DatasetCol]:
	query = f"""
		prefix cpmeta: <http://meta.icos-cp.eu/ontologies/cpmeta/>
		select ?colTitle ?valFormat ?isOptional ?isRegex ?flagColTitle where{{
			<{dataset_spec}> cpmeta:hasColumn ?col .
			?col cpmeta:hasColumnTitle ?colTitle ; cpmeta:hasValueFormat ?valFormat .
			optional{{?col cpmeta:isOptionalColumn ?isOptional}}
			optional{{?col cpmeta:isRegexColumn ?isRegex}}
			optional{{
				?flagCol cpmeta:isQualityFlagFor ?col ; cpmeta:hasColumnTitle ?flagColTitle .
				<{dataset_spec}> cpmeta:hasColumn ?flagCol .
			}}
		}}"""
	return [
		DatasetCol(
			col_title=as_string("colTitle", binding),
			val_format=as_uri("valFormat", binding),
			is_optional=as_opt_bool("isOptional", binding) or False,
			is_regex=as_opt_bool("isRegex", binding) or False,
			flag_col=as_opt_str("flagColTitle", binding)
		)
		for binding in meta.sparql_select(query).bindings
	]

def get_good_flags_per_spec(obj_specs: list[URI], meta: MetadataClient) -> dict[URI, list[str]]:
	uniq_obj_specs = [spec for spec in {spec for spec in obj_specs}]
	query = f"""
		prefix cpmeta: <http://meta.icos-cp.eu/ontologies/cpmeta/>
		select ?objFormat (group_concat(?goodFlag; separator=";") as ?goodFlags) where{{
			values ?spec {{ <{'> <'.join(uniq_obj_specs)}> }}
			?spec cpmeta:hasFormat ?objFormat .
			optional {{?objFormat cpmeta:hasGoodFlagValue ?goodFlag}}
		}}
		group by ?objFormat
	"""
	q_res = meta.sparql_select(query)
	joined = {
		as_uri("objFormat", binding): as_opt_str("goodFlags", binding)
		for binding in q_res.bindings
	}
	return {
		spec: flags.split(';')
		for spec, flags in joined.items()
		if flags
	}



import base64
import getpass
import json
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TypeAlias, Literal, Optional

from .envri import EnvriConfig
from .http import http_request

AuthSource: TypeAlias = Literal["Password", "Saml", "Orcid", "Facebook", "AtmoAccess"]
FreshnessMargin: timedelta = timedelta(hours = 1)
ExpiryMargin: timedelta = timedelta(seconds = 10)

@dataclass(frozen=True)
class AuthToken:
	user_id: str
	auth_source: AuthSource
	expiry_time: datetime
	cookie_value: str

	def _will_expire_in(self, delta: timedelta) -> bool:
		return (datetime.now() + delta) > self.expiry_time

	def is_expired(self) -> bool:
		return self._will_expire_in(ExpiryMargin)
	def is_fresh(self) -> bool:
		return not self._will_expire_in(FreshnessMargin)

def parse_auth_token(cookie_value: str) -> AuthToken:
	eq_idx = cookie_value.find("=")

	if eq_idx < 0 or eq_idx > 100:
		raise ValueError(f"Authentication token cookie value must contain '=' sign " +
			"close to the start of the string, but it was {cookie_value}")

	token_base64 = cookie_value[eq_idx + 1:]
	token_bin = base64.b64decode(token_base64)
	sep_idx = token_bin.index(ord('\x1e'))

	if sep_idx < 0:
		raise ValueError(f"Bad auth cookie format, missing expected separator byte 0x1e in: {token_base64}")

	ts_millis, user_id, auth_source = json.loads(token_bin[0:sep_idx].decode())

	expiry_time = datetime.fromtimestamp(ts_millis / 1000)
	return AuthToken(user_id, auth_source, expiry_time, cookie_value)

def fetch_auth_token(user_id: str, password: str, conf: EnvriConfig) -> AuthToken:
	url = conf.auth_service_base_url + "password/login"
	data = {'mail': user_id, 'password': password}
	headers = {"Content-Type": "application/x-www-form-urlencoded"}
	resp = http_request(url, "Fetching auth token", "POST", headers, data)
	cookie = resp.getheader("Set-Cookie")
	if cookie is None:
		raise Exception(f"Could not fetch auth token from {url}\nMissing value for header 'Set-Cookie'")
	else:
		cookie_value = cookie.split()[0]
		return parse_auth_token(cookie_value)

class AuthTokenProvider(ABC):
	@abstractmethod
	def get_token(self) -> AuthToken:
		pass

class TokenAuth(AuthTokenProvider):
	def __init__(self, token_or_cookie: AuthToken | str) -> None:
		if type(token_or_cookie) is AuthToken:
			self._token = token_or_cookie
		elif type(token_or_cookie) is str:
			self._token = parse_auth_token(token_or_cookie)
		else:
			raise ValueError(f"Bad argument to ConstAuthTokenProvider, expected AuthToken or the cookie value string")

	def get_token(self) -> AuthToken:
		t = self._token
		if t.is_expired():
			raise Exception(f"Authentication token expiration time is {t.expiry_time}, too late to use it")
		return t

class PasswordAuth(AuthTokenProvider):
	def __init__(self, user_id: str, password: str, conf: EnvriConfig, init_token: AuthToken | None = None) -> None:
		self._user_id = user_id
		self._password = password
		self._conf = conf
		self._token = init_token

	@property
	def has_fresh_token(self) -> bool:
		return (not not self._token) and self._token.is_fresh()

	def get_token(self) -> AuthToken:
		t = self._token
		if t and t.is_fresh():
			return t
		else:
			self._token = fetch_auth_token(self._user_id, self._password, self._conf)
			return self._token

	def save_to_file(self, file_path: str) -> None:
		data = {
			"user_id": self._user_id,
			"password": PasswordAuth.encode_password(self._password)
		}
		token = self._token
		if token:
			data["cookie_value"] = token.cookie_value
		with open(file_path, "w") as conf_file:
			json.dump(data, conf_file)

	@staticmethod
	def load_from_file(file_path: str, conf: EnvriConfig) -> "PasswordAuth":
		if not os.path.exists(file_path):
			docs_url = "https://github.com/ICOS-Carbon-Portal/data/tree/master/src/main/python/icoscp_core#authentication"
			raise Exception(f"Authentication config file not found at {file_path}. " +
				f"Please initialize authentication according to {docs_url} \n" +
				"(Remember to use Repository-specific imports in 'from icoscp_core.<repo> import auth')")
		with open(file_path, 'r') as conf_file:
			js = json.load(conf_file)
			user_id: str = js["user_id"]
			password: str = PasswordAuth.decode_password(js["password"])
			token: Optional[AuthToken] = None
			cookie_value: str | None = js.get("cookie_value")
			if cookie_value:
				token = parse_auth_token(cookie_value)
			return PasswordAuth(user_id, password, conf, token)

	@staticmethod
	def decode_password(encoded: str) -> str:
		return PasswordAuth._invert(bytes.fromhex(encoded)).decode()

	@staticmethod
	def encode_password(password: str) -> str:
		return PasswordAuth._invert(bytes(password, "utf-8")).hex()

	@staticmethod
	def _invert(barray: bytes) -> bytes:
		return bytes([~byte & 0xFF for byte in barray])

class ConfigFileAuth(AuthTokenProvider):
	def __init__(self, conf: EnvriConfig, conf_file_path: str | None = None) -> None:
		self._conf = conf
		if conf_file_path:
			self._conf_file_path = conf_file_path
		else:
			basedir = os.path.join(os.path.expanduser("~"), ".icoscp")
			os.makedirs(basedir, exist_ok = True)
			self._conf_file_path = os.path.join(basedir, conf.token_name + "_auth_conf.json")

		self._provider: PasswordAuth | None = None

	def get_token(self) -> AuthToken:
		inner = self._provider
		if not inner:
			inner = PasswordAuth.load_from_file(self._conf_file_path, self._conf)
			self._provider = inner

		has_fresh: bool = inner.has_fresh_token
		token = inner.get_token()
		if not has_fresh:
			inner.save_to_file(self._conf_file_path)
		return token

	def init_config_file(self) -> None:
		user_id = input("Enter your username: ")
		password = getpass.getpass("Enter your password: ")
		inner = PasswordAuth(user_id, password, self._conf)
		inner.save_to_file(self._conf_file_path)
		self._provider = inner



from .envri import ICOS_CONFIG
from .bootstrap import Bootstrap
from .sparql import as_string, as_uri
from .metacore import URI

bootstrap = Bootstrap(ICOS_CONFIG)
auth, meta, data = bootstrap.fromPasswordFile()

ONTOLOGY_PREFIX: str = "http://meta.icos-cp.eu/ontologies/cpmeta/"

def onto_uri(suffix: str) -> str:
	return ONTOLOGY_PREFIX + suffix

ATMO_STATION  = onto_uri("AS")
ECO_STATION   = onto_uri("ES")
OCEAN_STATION = onto_uri("OS")


_station_class_lookup: dict[URI, str] | None = None

def station_class_lookup() -> dict[URI, str]:
	"""
	ICOS-specific utility function producing lookup of station class by
	station URI. Caches the result after the first call. The cache can
	be dropped by setting `_station_class_lookup` to `None`

	:returns: dictionary from station URI id to the station class string
	"""
	global _station_class_lookup
	if _station_class_lookup is not None:
		return _station_class_lookup
	statClassQuery = f"select * where{{?station <{onto_uri('hasStationClass')}> ?class}}"
	_station_class_lookup = {
		as_uri('station', row): as_string('class', row)
			for row in meta.sparql_select(statClassQuery).bindings
	}
	return _station_class_lookup



from typing import TypeAlias, Any
from dataclasses import dataclass
from .metacore import (
	FeatureCollection,
	Position,
	LatLonBox,
	GeoTrack,
	Polygon,
	Circle,
	Pin
)

@dataclass(frozen=True)
class GeoJsonHolder:
	geo: dict[str, Any]

@dataclass(frozen=True)
class FeatureCollectionWithGeo(FeatureCollection, GeoJsonHolder): pass

@dataclass(frozen=True)
class PositionWithGeo(Position, GeoJsonHolder): pass

@dataclass(frozen=True)
class LatLonBoxWithGeo(LatLonBox, GeoJsonHolder): pass

@dataclass(frozen=True)
class GeoTrackWithGeo(GeoTrack, GeoJsonHolder): pass

@dataclass(frozen=True)
class PolygonWithGeo(Polygon, GeoJsonHolder): pass

@dataclass(frozen=True)
class CircleWithGeo(Circle, GeoJsonHolder): pass

@dataclass(frozen=True)
class PinWithGeo(Pin, GeoJsonHolder): pass

GeoFeatureWithGeo: TypeAlias = (
	FeatureCollectionWithGeo |
	PositionWithGeo |
	LatLonBoxWithGeo |
	GeoTrackWithGeo |
	PolygonWithGeo |
	CircleWithGeo |
	PinWithGeo
)

@dataclass
class Point:
	"""
	Geographical point

	Attributes:
		`lat` (float): WGS-84 latitude
		`lon` (float): WGS-84 longitude
	"""
	lat: float
	lon: float



from datetime import datetime
from typing import TypeAlias, TypeVar, Optional, Any, Callable
from dataclasses import dataclass
import re
import json

from .http import http_request

@dataclass(frozen=True)
class BoundUri:
	uri: str

@dataclass(frozen=True)
class BoundLiteral:
	value: str
	datatype: Optional[str]

BoundValue: TypeAlias = BoundUri | BoundLiteral

Binding: TypeAlias = dict[str, BoundValue]

@dataclass(frozen=True)
class SparqlResults:
	variable_names: list[str]
	bindings: list[Binding]

def as_uri(varname: str, binding: Binding) -> str:
	bv = _lookup(varname, binding)
	if type(bv) is BoundUri:
		return bv.uri
	else:
		raise _type_error(varname, "a uri value", bv)

ResT = TypeVar('ResT')
SparqlVarParser: TypeAlias = Callable[[str, Binding], ResT]

def _as_type(xsd_type: str | None, parser: Callable[[str], ResT]) -> SparqlVarParser[ResT]:
	return lambda varname, binding:(
		parser(_lookup_literal_value(varname, xsd_type, binding))
	)

def _as_opt(plain: SparqlVarParser[ResT]) -> SparqlVarParser[ResT | None]:
	return lambda varname, binding: plain(varname, binding) if varname in binding.keys() else None

as_bool: SparqlVarParser[bool] = _as_type("boolean", lambda s: True if s.lower() == "true" else False)
as_int: SparqlVarParser[int] = _as_type("integer", lambda s: int(s))
as_double: SparqlVarParser[float] = _as_type("double", lambda s: float(s))
as_float: SparqlVarParser[float] = _as_type("float", lambda s: float(s))
as_long: SparqlVarParser[int] = _as_type("long", lambda s: int(s))
as_string: SparqlVarParser[str] = _as_type(None, lambda s: s)
as_datetime: SparqlVarParser[datetime] = _as_type(
	xsd_type = "dateTime",
	parser = lambda s: datetime.fromisoformat(re.sub(r'Z$', '+00:00', s))
)
as_opt_bool = _as_opt(as_bool)
as_opt_double = _as_opt(as_double)
as_opt_float = _as_opt(as_float)
as_opt_str = _as_opt(as_string)
as_opt_uri = _as_opt(as_uri)

def _lookup(varname: str, binding: Binding) -> BoundValue:
	v = binding.get(varname)
	if not v:
		raise ValueError(f"Variable {varname} had no bound values in SPARQL response")
	else:
		return v

def _lookup_literal_value(varname: str, datatype: Optional[str], binding: Binding) -> str:
	bv = _lookup(varname, binding)
	if type(bv) is BoundLiteral:
		if not(datatype) or bv.datatype == ("http://www.w3.org/2001/XMLSchema#" + datatype):
			return bv.value
		else:
			raise _type_error(varname, f"literal datatype {datatype}", bv)
	else:
		raise _type_error(varname, "a literal", bv)

def _type_error(varname: str, expected: str, bv: BoundValue) -> ValueError:
	msg = f"Was expecting {expected}, got value {bv} for variable {varname} in SPARQL results"
	return ValueError(msg)

def get_sparql_select_json(endpoint: str, query: str, disable_cache: bool) -> Any:
	headers: dict[str, str] = {"Accept": "application/json"}
	if disable_cache:
		headers["Cache-Control"] = "no-cache"
		headers["Pragma"] = "no-cache"
	resp = http_request(endpoint, "SPARQL SELECT", "POST", headers, query)
	return json.loads(resp.read())

def sparql_select(endpoint: str, query: str, disable_cache: bool) -> SparqlResults:
	js = get_sparql_select_json(endpoint, query, disable_cache)
	varnames: list[str] = js["head"]["vars"]
	binding_dicts: list[dict[str, Any]] = js["results"]["bindings"]
	bindings = [{key: _parse_bound_value(v) for key, v in b_dict.items()} for b_dict in binding_dicts]
	return SparqlResults(varnames, bindings)

def _parse_bound_value(v: Any) -> BoundValue:
	match v["type"]:
		case "uri": return BoundUri(uri = v["value"])
		case "literal": return BoundLiteral(value = v["value"], datatype=v.get("datatype"))
		case "bnode": raise ValueError("Blank nodes in SPARQL return results are not supported by this" +
			" library. Please re-formulate your query.")
		case bad: raise ValueError(f"Unexpected result type encountered in SPARQL response: {bad}")
